from LMS.Message.Definitions.LMS_FieldMap import LMS_FieldMap
from LMS.Message.Definitions.LMS_MessageText import LMS_MessageText


class MSBTEntry:
    """A class that represents an entry in a MSBT file."""

    def __init__(
        self,
        name: str,
        message: str | LMS_MessageText,
        attribute: LMS_FieldMap | bytes = None,
        style_index: int = None,
    ):
        self.name = name

        if isinstance(message, str):
            self._message = LMS_MessageText(message)
        else:
            self._message = message

        self._attribute = attribute
        self.style_index = style_index

    @property
    def message(self) -> LMS_MessageText:
        """The message object for the instance."""
        return self._message

    @property
    def attribute(self) -> LMS_FieldMap | bytes:
        """The attribute for the instance."""
        return self._attribute



from __future__ import annotations

from dataclasses import dataclass

from LMS.Config.Definitions.Value import ValueDefinition


@dataclass(frozen=True)
class TagDefinition:
    """Class that represents a signle tag definition in the structure."""

    group_name: str
    group_index: int
    tag_name: str
    tag_index: int
    description: str
    parameters: list[ValueDefinition] | None


@dataclass(frozen=True)
class TagConfig:
    """Class that represents an tag structure definition."""

    group_map: dict[int, str]
    definitions: list[TagDefinition]

    def get_definition_by_names(self, group: str, tag: str) -> TagDefinition:
        group_index = None
        for i, name in self.group_map.items():
            if name == group:
                group_index = i
                break

        if group_index is None:
            raise KeyError(f"Group name '{group}' was not found! Is the group defined?")

        for tag_def in self.definitions:
            if tag_def.group_index == group_index and tag_def.tag_name == tag:
                return tag_def

        raise KeyError(
            f"Tag name '{tag}' not found in group '{group}'. Os the tag defined?"
        )

    def get_definition_by_indexes(self, group: int, tag: int) -> TagDefinition:
        for tag_def in self.definitions:
            if tag_def.group_index == group and tag_def.tag_index == tag:
                return tag_def

        raise KeyError(
            f"Tag index of {tag} was not found in group index {group}. Is the tag defined?"
        )



import importlib.resources as pkg_resources
import os
from typing import Self

import yaml

from LMS.Config.Definitions.Attributes import AttributeConfig
from LMS.Config.Definitions.Tags import TagConfig, TagDefinition
from LMS.Config.Definitions.Value import ValueDefinition
from LMS.Field.LMS_DataType import LMS_DataType
from LMS.Project.MSBP import MSBP


class TitleConfig:
    """Represents a configuration for a specific game/title."""

    # Populate the preset list from the directory
    PRESET_LIST = [
        name.removesuffix(".yaml") for name in pkg_resources.contents("LMS.Config.Presets") if name.endswith(".yaml")
    ]

    def __init__(self, attribute_configs: dict[str, AttributeConfig], tag_config: TagConfig):
        self._attribute_configs = attribute_configs
        self._tag_config = tag_config

    @property
    def attribute_configs(self) -> dict[str, AttributeConfig]:
        """The map of attribute config instances."""
        return self._attribute_configs

    @property
    def tag_config(self) -> TagConfig:
        """The loaded tag config instance."""
        return self._tag_config

    @classmethod
    def load_preset(cls, game: str) -> Self:
        """Loads an existing preset from the package."""
        if game not in cls.PRESET_LIST:
            raise FileNotFoundError(f"Preset '{game}' not found.")

        with pkg_resources.open_text("LMS.Config.Presets", f"{game}.yaml") as f:
            return cls.load_config(f.read())

    @classmethod
    def load_file(self, file_path: str) -> Self:
        """Loads a config from a file.

        :param file_path: the path to the config."""
        with open(file_path, "r") as f:
            return TitleConfig.load_config(f.read())

    @classmethod
    def load_config(cls, content: str | dict) -> Self:
        """Loads the config of a specified game.

        :param content: the config content, as a string or loaded as a dictionary."""
        if isinstance(content, str):
            parsed_content = yaml.safe_load(content)
        else:
            parsed_content = content

        group_map = {0: "System"}
        tag_config = []

        # Add the System tag definitions to the config
        with open(r"LMS\Message\Tag\System.yaml", "r") as f:
            system_tags = yaml.safe_load(f)
            tag_config = system_tags

        # Combine with the rest of the cofnig
        group_map |= parsed_content["tag_definitions"]["groups"]
        for tag_def in parsed_content["tag_definitions"]["tags"]:
            tag_config.append(tag_def)

        # Load tag definitions
        tag_definitions = []
        for tag_def in tag_config:
            tag_name = tag_def["name"]
            group_index, tag_index = tag_def["group_index"], tag_def["tag_index"]
            group_name = group_map[group_index]
            tag_description = tag_def["description"]

            group_name = group_map[group_index]

            parameters = None
            if "parameters" in tag_def:
                parameters = []
                for param_def in tag_def["parameters"]:
                    param_name = param_def["name"]
                    param_description = param_def["description"]
                    datatype = LMS_DataType.from_string(param_def["datatype"])
                    list_items = param_def.get("list_items")
                    parameters.append(ValueDefinition(param_name, param_description, datatype, list_items))

            tag_definitions.append(
                TagDefinition(
                    group_name,
                    group_index,
                    tag_name,
                    tag_index,
                    tag_description,
                    parameters,
                )
            )

        tag_config = TagConfig(group_map, tag_definitions)

        # Load the attribute definitions
        attribute_config = {}
        for structure in parsed_content["attributes"]:
            structure_name = structure["name"]

            definitions = []
            for info in structure["definitions"]:
                name, description = info["name"], info["description"]
                datatype = LMS_DataType.from_string(info["datatype"])
                list_items = info.get("list_items")

                definition = ValueDefinition(name, description, datatype, list_items)
                definitions.append(definition)

            structure = AttributeConfig(structure_name, structure["description"], definitions)
            attribute_config[structure_name] = structure

        return cls(attribute_config, tag_config)

    @staticmethod
    def generate_file(file_path: str, project: MSBP) -> None:
        with open(file_path, "w+") as f:
            yaml.safe_dump(
                TitleConfig.generate_profile(project),
                f,
                default_flow_style=False,
                sort_keys=False,
            )

    @staticmethod
    def generate_config(project: MSBP) -> dict | None:
        """Creates a message config file for the game.

        :param project: a MSBP object."""
        # TODO: Add custom node definitions

        tag_key = "tag_definitions"
        attr_key = "attributes"

        config = {}
        # Source files may be a path to a non-existent directory.
        # These files were generated from the source machine from the actual libMS tool
        # Shorten the filename with basname and replace the extension with .msbt for lookup later when reading a MSBT
        config[tag_key] = {
            "groups": {i + 1: group.name for i, group in enumerate(project.tag_groups[1:])},
            "tags": [],
        }

        # Slice to exclude System group
        for group_i, group in enumerate(project.tag_groups[1:], start=1):
            for tag_i, info in enumerate(group.tag_definitions):

                definition = {
                    "name": info.name,
                    "group_index": group_i,
                    "tag_index": tag_i,
                    "description": "",
                    "parameters": [],
                }

                for param_info in info.param_info:
                    param_definition = {
                        "name": param_info.name,
                        "description": "",
                        "datatype": param_info.datatype.to_string(),
                    }

                    if param_info.datatype is LMS_DataType.LIST:
                        param_definition["list_items"] = param_info.list_items

                    definition["parameters"].append(param_definition)

                config[tag_key]["tags"].append(definition)

        # Set main attribute entries as the definitions from the MSBP
        # Since most games use one MSBP these act the primary definitions
        config[attr_key] = []

        attr_definitions = []
        for attr_info in project.attribute_info:
            definition = {
                "name": attr_info.name,
                "description": "",
                "datatype": attr_info.datatype.to_string(),
            }
            if attr_info.datatype is LMS_DataType.LIST:
                definition["list_items"] = attr_info.list_items

            attr_definitions.append(definition)

        # Main attribute entries
        config[attr_key].append(
            {
                "name": project.name,
                "description": "",
                "definitions": attr_definitions,
            }
        )

        return config



from dataclasses import dataclass

from LMS.Field.LMS_DataType import LMS_DataType


@dataclass(frozen=True)
class ValueDefinition:
    name: str
    description: str
    datatype: LMS_DataType
    list_items: dict[int, str]



from dataclasses import dataclass, field

from LMS.Config.Definitions.Value import ValueDefinition


@dataclass(frozen=True)
class AttributeConfig:
    """Class that represents an attribute structure definition.

    Configs may have multiple, linked to the if a game has multiple MSBP files."""

    name: str
    description: str = field(repr=False)
    definitions: list[ValueDefinition]



from LMS.Config.Definitions.Attributes import AttributeConfig
from LMS.Field.LMS_DataType import LMS_DataType
from LMS.Field.LMS_Field import LMS_DataType, LMS_Field
from LMS.Field.Stream import read_field, write_field
from LMS.FileIO.Stream import FileReader, FileWriter
from LMS.Message.Definitions.LMS_FieldMap import LMS_FieldMap


def read_encoded_atr1(
    reader: FileReader, section_size: int
) -> tuple[list[bytes], int, list[str]]:
    absolute_size = section_size + reader.tell()

    attribute_count = reader.read_uint32()
    size_per_attribute = reader.read_uint32()

    attributes = [reader.read_bytes(size_per_attribute) for _ in range(attribute_count)]
    string_table = None

    if section_size > 8 + size_per_attribute * attribute_count:
        string_table = reader.read_bytes(absolute_size - reader.tell())

    return (attributes, size_per_attribute, string_table)


def write_encoded_atr1(
    writer: FileWriter,
    attributes: list[bytes],
    size_per_attribute: int,
    string_table: bytes | None,
):
    writer.write_uint32(len(attributes))

    if attributes:
        writer.write_uint32(size_per_attribute)
    else:
        writer.write_uint32(0)

    for attr in attributes:
        writer.write_bytes(attr)

    if string_table is not None:
        writer.write_bytes(string_table)


def read_decoded_atr1(
    reader: FileReader, structure: AttributeConfig
) -> tuple[list[LMS_FieldMap], int, None]:
    section_start = reader.tell()

    attr_count = reader.read_uint32()
    size_per_attr = reader.read_uint32()

    attributes = []
    attr_start = reader.tell()

    for _ in range(attr_count):
        reader.seek(attr_start)

        attribute = {}
        for definition in structure.definitions:
            if definition.datatype is LMS_DataType.STRING:
                last = reader.tell() + 4
                reader.seek(section_start + reader.read_uint32())
                value = LMS_Field(reader.read_str_variable_encoding(), definition)

                reader.seek(last)
            else:
                value = read_field(reader, definition)

            attribute[definition.name] = value

        attributes.append(attribute)

        # Move to the start of the next attribute
        attr_start += size_per_attr

    return attributes, size_per_attr, None


def write_decoded_atr1(
    writer: FileWriter, attributes: list[LMS_FieldMap], size_per_attribute: int
) -> None:
    writer.write_uint32(len(attributes))
    writer.write_uint32(size_per_attribute)

    string_table = []
    string_offset = 8 + size_per_attribute * len(attributes)
    for attr in attributes:
        for field in attr.values():
            if field.datatype is not LMS_DataType.STRING:
                write_field(writer, field)
            else:
                writer.write_uint32(string_offset)
                string_offset += (
                    len(field.value) * writer.encoding.width
                ) + writer.encoding.width
                string_table.append(field.value)

    for string in string_table:
        writer.write_variable_encoding_string(string)



from LMS.Common.LMS_FileInfo import LMS_FileInfo
from LMS.Message.MSBTEntry import MSBTEntry


class MSBT:
    """A Class that represents a MSBT file.

    https://nintendo-formats.com/libs/lms/msbt.html."""

    def __init__(self, info: LMS_FileInfo):
        self._info = info

        self._entries: list[MSBTEntry] = []

        self.size_per_attribute = 0

        # 101 is default for almost all games. However the value can be overriden by some games (i.e Echos of Wisdom).
        # Due to this, the slot count is set dynamically when LBL1 is read.
        self.slot_count = None

        self.attr_string_table: bytes = None
        self.encoded_attributes = True

        # List of section names to preserve order
        self.section_list: list[str] = []

        # List of unsupported sections mapped to their raw data
        self.unsupported_sections: dict[str, bytes] = {}

    @property
    def info(self) -> LMS_FileInfo:
        """The file info for the MSBT instance."""
        return self._info

    @property
    def entries(self) -> list[MSBTEntry]:
        """The list of entries for the MSBT instance."""
        return self._entries

    def add_entry(
        self,
        name: str,
        text: str = None,
        attribute: dict[str, int | str | float | bool | bytes] = None,
        style_index: int = None,
    ) -> None:
        """Adds an entry to the MSBT instance.

        :param name: the name of the entry."""
        if name in [entry.name for entry in self.entries]:
            raise KeyError(f"The label '{name}' already exists!")

        self._entries.append(MSBTEntry(name, text, attribute, style_index))

    def get_entry(self, name: str) -> MSBTEntry:
        """Retrieves an entry given it's name.

        :param name: the name of the entry."""
        for entry in self._entries:
            if name == entry.name:
                return entry

        raise KeyError(f"The label '{name}' does not exist!")
